# adding_up.py: JF Bouchaudy v0.5 2025-03-05
# ------------------------------------------------------
# methode pour trouver le cablage inverse du rotor rapide
# à partir d'un message long et du clair correspondant
# -------------------------------------------------------
# Ref: F. Carter, Alan Turing, John Wright
# -------------------------------------------------------
# Le coeur de l'algorithme: 
# recherche de deductions ou une seule lettre est nouvelle
# est on se base (si possible) uniquement sur les confirm-
# ations. 
# -------------------------------------------------------
# on saisie un fichier (sur STDIN), on double chaque ligne
# on saisie (en argument?) les lettres deja connues
# si on saisie une chaine sans lettres (ex: "....")
# on recherche "from Scratch": on WX.Y..Z on teste toutes
# les lettres possibles: 26x25x24x23 ~= 400,000 pos.
# -------------------------------------------------------
# Ne serais que pour des tests, on donne en argument une
# ou plusieurs listes de lettres avec ecart 
# ex: A....G...F,K....D  Le G est separé de 4 lettres de A
# -------------------------------------------------------
# v0.4: la version publiée: fichier de données avec turnover
# v0.5: levée d'Exception si solution est impossible
# -------------------------------------------------------
import string
import sys
import os

# --------------------------------------------------------
class IMPOSSIBLE(Exception):
    pass
# --------------------------------------------------------
def fait_deduc(c1, c2):
    resultat = []
    letA1 = c1[0]; letA2 = c1[1]
    letB1 = c2[0]; letB2 = c2[1]
    let_qua = c1 + c2
    nb = 0
    for let in ENS_LET:
        nb += let_qua.count(let)
    if nb == 3:
        deduc1 = deduc2 = deduc3 = deduc4 = ""
        if   letA1 == letB1:
            deduc1 = letA1 , letB2 
            deduc2 = letA2 , letB1
        elif letA1 == letB2:
            deduc1 = letA1 , letB1
            deduc2 = letA2 , letB2
        elif letA2 == letB1:
            deduc1 = letA1 , letB1
            deduc2 = letA2 , letB2
        elif letA2 == letB2:
            deduc1 = letA1 , letB2
            deduc2 = letA2 , letB1
        else:
            deduc1 = letA1 , letB1
            deduc2 = letA2 , letB2
            deduc3 = letA1 , letB2
            deduc4 = letA2 , letB1
        if deduc1 != "" and deduc3 == "":
            res = valide(deduc1, ECART)
            resultat.append([deduc1,ECART,res])
            res = valide(deduc2, ECART)
            resultat.append([deduc2,ECART,res])
        if deduc3 != "":
            res = valide(deduc1, ECART)
            resultat.append([deduc1,ECART,res])
            res = valide(deduc2, ECART)
            resultat.append([deduc2,ECART,res])
            res = valide(deduc3, ECART)
            resultat.append([deduc3,ECART,res])
            res = valide(deduc4, ECART)
            resultat.append([deduc4,ECART,res])
    return resultat

#--------------------------------------------------------
def cree_dico():
  for LETTRES in LES_LETTRES:
    for car in LETTRES:
        if car in ALPHA:
            DICO[car] = {}
            pos_car = LETTRES.find(car)
            for i in range(len(LETTRES)):
                if LETTRES[i] in ALPHA and LETTRES[i] != car:
                    position = i - pos_car
                    if position < 0:
                        position = 26 - abs(position)
                    DICO[car][position] = LETTRES[i]

#--------------------------------------------------------
def valide(un_couple, ecart):
    if un_couple[0] in DICO:
        if ecart in DICO[un_couple[0]]:
            if DICO[un_couple[0]][ecart] == un_couple[1]:
                return 1
            else:
                return -1
        else:
            if un_couple[1] in DICO[un_couple[0]].values():
                return -2
    return 0
        
#--------------------------------------------------------
# une_chaine: ex:   "A.....M" =>  [('A','M'),6, ...]
# genere une nouvelle hypothese sous forme de chaine
# par exemple "A...M...G" +  "A..D" => "A..DM...G"
# ====> il faudra recreer le DICO !!
def modif_dico(tb):
    global ENS_LET
    letG = tb[0][0]
    letD = tb[0][1]
    ecart= tb[1]
    trouve = 0
    num_modif = -1
    #--- on recherche la chaine ou se trouve une
    #    ou l'autre lettre [on verra plus tard si les 2]
    for i in range(len(LES_LETTRES)):
        L = LES_LETTRES[i]
        if letG in L and letD in L:
            return
        elif letG in L and letD not in L:
            trouve += 1
            num_modif = i
            couple = (letG, letD)
        elif letD in L and letG not in L:
            trouve += 1
            num_modif = i
            couple = (letG, letD)
        else:
            pass
    if trouve == 0:
        #--- on ajouter une liste
        LES_LETTRES.append(letG + ("."*(ecart-1)) + letD + ("."*(24-(ecart-1))))
    elif trouve == 1:
        #--- on ajoute à une liste existante
        L = LES_LETTRES[num_modif]
        posG = L.find(couple[0])
        posD = L.find(couple[1])
        if posG >= 0 and posD >= 0:
            return
        if posG >= 0 and posD < 0:
            posFinal = (posG + ecart) % 26
            if L[posFinal] != ".":
                raise IMPOSSIBLE()
            L = L[:posFinal] + letD + L[posFinal+1:]
        elif posD >= 0 and posG < 0:
            posFinal = (posD - ecart) % 26
            if L[posFinal] != ".":
                raise IMPOSSIBLE()
            L = L[:posFinal] + letG + L[posFinal+1:]
        LES_LETTRES[num_modif] = L
        pass
    elif trouve == 2:
        #--- On fusionne deux listes 
        #--- 1) on calcule les positions
        pos_gauche = num_gauche = -1
        pos_droite = num_droite = -1
        for i in range(len(LES_LETTRES)):
            L = LES_LETTRES[i]
            if letG in L:
                num_gauche = i
                pos_gauche = L.find(letG)
            if letD in L:
                num_droite = i
                pos_droite = L.find(letD)
        ecart_courant = pos_droite - pos_gauche
        #--- 2) on decale la liste auquel appartient la letD
        L = LES_LETTRES[num_droite]
        decal = ecart - ecart_courant
        if decal > 0:
            x = L[26-decal:] + L[:26-decal]
        elif decal < 0:
            x = L[abs(decal):] + L[:abs(decal)] 
        elif decal == 0:
            x = L
        y = LES_LETTRES[num_gauche]
        print("--->", x, num_droite, ecart, ecart_courant, decal, len(x))
        pass
        print("G : ", LES_LETTRES[num_gauche])
        print("D : ", LES_LETTRES[num_droite])
        print("Fusion", letG, letD, ecart, ecart_courant, decal, end="," )
        print(num_gauche, pos_gauche, num_droite, pos_droite)
        #--- 3) On recopie la liste decalee dans l'autre
        tb = ["."]*26
        for i in range(26):
            if   x[i] in ALPHA:
                tb[i] = x[i]
            elif y[i] in ALPHA:
                tb[i] = y[i]
        z = "".join(tb)
        #--- 4) On supprime la liste qui a ete decalle
        LES_LETTRES[num_gauche] = z
        LES_LETTRES.pop(num_droite)
    ENS_LET = set("".join(LES_LETTRES)) & set(ALPHA)

#--------------------------------------------------------
def aff_let():
    for LETTRES in LES_LETTRES:
        print("===>", LETTRES)

#--------------------------------------------------------
def aff_dico():
    print("Lettres connues: ", ENS_LET)
    for LETTRES in LES_LETTRES:
        print("=>", LETTRES)
    print("Dico: ")
    for cle in DICO.keys():
        print(cle, "=>", DICO[cle], ":", DICO[cle].values())
#========================================================
DEBUG = os.getenv("DEBUG")

ALPHA = string.ascii_uppercase

chaine = sys.argv[1]
LES_LETTRES = chaine.split(",")
#--- on complete les listes à 26 caracteres
for i in range(len(LES_LETTRES)):
    LETTRES = LES_LETTRES[i]
    if len(LETTRES) < 26:
        LETTRES = LETTRES + "."*(26-len(LETTRES))
        LES_LETTRES[i] = LETTRES

DICO = {}
cree_dico()

ENS_LET = set("".join(LES_LETTRES)) & set(ALPHA)
if DEBUG:
    print("Lettres connues: ", ENS_LET)

fic = sys.argv[2]
fic_des = open(fic)
LES_LIGNES = []
fic = fic_des.readlines()
fic_des.close()

for ligne in fic:
    ligne = ligne.rstrip()
    couples = ligne.split(":")
    if len(couples) < 26:
        LES_LIGNES.append( list(couples) )
    else:
        LES_LIGNES.append( list(couples + couples) )

MAX_ECART = 13
aff_let()
#-------------------------------------------
for HYPO in range(30):
 try:
  NEW_DEDUC = False
  PILE = []
  for ECART in range(1,MAX_ECART):
    for i in range(len(LES_LIGNES)):
        la_ligne = LES_LIGNES[i]
        for j in range(26):
            if j >= len(LES_LIGNES[i]) or (j+ECART) >= len(LES_LIGNES[i]):
                continue
            c1 = la_ligne[j]
            c2 = la_ligne[j+ECART]
            resu = fait_deduc(c1,c2)
            if len(resu) > 0:
                for k in range(len(resu)):
                    if resu[k][2] > 0:
                        print("  Confirmation: ", c1, c2, end="")
                        if k == 0 or k == 2:
                            if resu[k][2] == 1 or  \
                               resu[k+1][2] == 1:
                                if resu[k][2] == 1:
                                    if resu[k+1][2] != 1:
                                        PILE.append(resu[k+1])
                                        NEW_DEDUC = True
                                        (res_a, res_b), res_c, res_d = resu[k+1]
                                        print(", New: %1s-%02d-%1s" % (res_a, res_c, res_b))
                        elif k == 1 or k == 3:
                            if resu[k][2] == 1 or  \
                               resu[k-1][2] == 1:
                                if resu[k][2] == 1:
                                    if resu[k-1][2] != 1:
                                        PILE.append(resu[k-1])
                                        NEW_DEDUC = True
                                        (res_a, res_b), res_c, res_d = resu[k-1]
                                        print(", New: %1s-%02d-%1s" % (res_a, res_c, res_b))
                        else:
                            if DEBUG:
                                print("double confir")
  if NEW_DEDUC:
    print("New hypothesis:")
    for un_elt in PILE:
        modif_dico(un_elt) 
    aff_let()
    cree_dico()
  else:
    print("<<< No new hypothesis >>>")
    aff_let()
    break
  input("? ")
 except IMPOSSIBLE:
    print("<<< IMPOSSIBLE >>>")
    break

